(*
Gabriel Corneanu

A true (ADO only!!) WideMemo blob field. It means it's descendant of TBlob field
(actualy from TMemo) and it properly gets it's data. No UTF8 or anything else,
directly Unicode.
All you need to do is to include this unit. If using TNT framework (db controls),
you should get directly proper Unicode display.

It only makes sense to use with LongVarWChar type, not with LongVarChar because
it will always convert to Unicode.

TWideStringADOBlobStream is a blob stream that uses WideString as native storage

TWBADODataSet is a TADODataSet descendant that creates a proper blob stream
(as a WideString if necessary).
It can be used as it is, or as a reference

Limitations:
- LoadFromStream/SaveToStream will work only if the dataset is TWBADODataSet.
*)
unit TntWideADOMemoField;

interface
uses
  ADOInt,
  ADODB,
  Classes,
  DB,
  TntDB;

type
  TWideADOMemoField = class(TMemoField, IWideStringField)
  private
    function  GetAsWideString: WideString;
    procedure SetAsWideString(const Value: WideString);
    function  GetWideDisplayText: WideString;
    function  GetWideEditText: WideString;
    procedure SetWideEditText(const Value: WideString);
  protected
    function  GetSize: Integer; override;
    procedure GetText(var Text: String; DisplayText: Boolean); override;
  public
    constructor Create(AOwner: TComponent); override;
  end;

  //inspired from TAdoBlobStream
  //if data is a WideString, preserve (or convert to WideString)
  TWideStringADOBlobStream = class(TMemoryStream)
  private
    FField: TBlobField;
    FDataSet: TCustomADODataSet;
    FBuffer: PChar;
    FFieldNo: Integer;
    FModified: Boolean;
    FData: Variant;
    FFieldData: Variant;
  protected
    procedure ReadBlobData;
    function  Realloc(var NewCapacity: Longint): Pointer; override;
  public
    constructor Create(Field: TBlobField; Mode: TBlobStreamMode);
    destructor  Destroy; override;
    function    Write(const Buffer; Count: Longint): Longint; override;
    procedure   Truncate;
  end;


  //class that creates a proper blob stream
  TWBADODataSet = class(TADODataSet)
  public
    function CreateBlobStream(Field: TField; Mode: TBlobStreamMode): TStream; override;
  end;

implementation

uses
  DBConsts,
  Variants;


type
  TAccessCustomADODataSet = class(TCustomADODataSet);

{ TWideADOMemoField }

constructor TWideADOMemoField.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
end;

function TWideADOMemoField.GetWideDisplayText: WideString;
begin
  Result := TntWideStringField_GetWideDisplayText(Self, nil);
end;

function TWideADOMemoField.GetWideEditText: WideString;
begin
  Result := TntWideStringField_GetWideEditText(Self, nil);
end;

function TWideADOMemoField.GetAsWideString: WideString;
var
  Len: Integer;
begin
  with TWideStringADOBlobStream.Create(Self, bmRead) do
    try
      Len := Size;
      SetLength(Result, Len div SizeOf(WideChar));
      ReadBuffer(Pointer(Result)^, Len);
    finally
      Free;
    end;
end;

procedure TWideADOMemoField.SetAsWideString(const Value: WideString);
begin
  with TWideStringADOBlobStream.Create(Self, bmWrite) do
    try
      WriteBuffer(Pointer(Value)^, Length(Value) * SizeOf(WideChar));
    finally
      Free;
    end;
end;

procedure TWideADOMemoField.SetWideEditText(const Value: WideString);
begin
  TntWideStringField_SetWideEditText(Self, nil , Value);
end;

function TWideADOMemoField.GetSize: Integer;
begin
  with TWideStringADOBlobStream.Create(Self, bmRead) do
    try
      Result := Size;
    finally
      Free;
    end;
end;

procedure TWideADOMemoField.GetText(var Text: String; DisplayText: Boolean);
begin
//this looses Unicode, but will return the text instead of "(Memo)"
  if DisplayText then Text := GetWideDisplayText
                 else Text := GetWideEditText;
end;



{ TPreserveWideStringADOBlobStream }

constructor TWideStringADOBlobStream.Create(Field: TBlobField; Mode: TBlobStreamMode);
begin
  FField := Field;
  FFieldNo := FField.FieldNo - 1;
  FDataSet := TAccessCustomADODataSet(FField.DataSet as TCustomADODataSet);
  FFieldData := Null;
  FData := Null;
  if not TAccessCustomADODataSet(FDataSet).GetActiveRecBuf(FBuffer) then Exit;
  if Mode <> bmRead then
  begin
    if FField.ReadOnly then
      DatabaseErrorFmt(SFieldReadOnly, [FField.DisplayName], FDataSet);
    if not (FDataSet.State in [dsEdit, dsInsert]) then
      DatabaseError(SNotEditing, FDataSet);
  end;
  if Mode = bmWrite then Truncate
  else ReadBlobData;
end;

destructor TWideStringADOBlobStream.Destroy;
begin
  if FModified then
  try
    TAccessCustomADODataSet(FDataSet).SetFieldData(FField, @FData);
    FField.Modified := True;
    TAccessCustomADODataSet(FDataSet).DataEvent(deFieldChange, Longint(FField));
  except
    ApplicationHandleException(Self);
  end;
  inherited Destroy;
end;

procedure TWideStringADOBlobStream.ReadBlobData;
begin
  FDataSet.GetFieldData(FField, @FFieldData, True);
  if not VarIsNull(FFieldData) then
  begin
    if VarIsStr(FFieldData) then
    //if VarType(FFieldData) = varOleStr then
    begin
      //GC: convert to WideString if AnsiString!!!
      //FFieldData := string(FFieldData);
      if VarType(FFieldData) <> varOleStr then
        FFieldData := WideString(FFieldData);
      Size := Length(FFieldData) * SizeOf(WideChar);
    end else
      Size := VarArrayHighBound(FFieldData, 1) + 1;
    FFieldData := Null;
  end;
end;

function TWideStringADOBlobStream.Realloc(var NewCapacity: Longint): Pointer;

  procedure VarAlloc(var V: Variant; StrField: Boolean);
  var
    S: WideString;
  begin
    if StrField then
    begin
      if not VarIsNull(V) then S := WideString(V);
      SetLength(S, NewCapacity div SizeOf(WideChar));
      V := S;
    end else
    begin
      if VarIsClear(V) or VarIsNull(V) then
        V := VarArrayCreate([0, NewCapacity-1], varByte) else
        VarArrayRedim(V, NewCapacity-1);
    end;
  end;

begin
  Result := Memory;
  if NewCapacity <> Capacity then
  begin
    if VarIsArray(FData) then VarArrayUnlock(FData);
    if NewCapacity = 0 then
    begin
      FData := Null;
      Result := nil;
    end else
    begin
      if VarIsNull(FFieldData) then
        VarAlloc(FData, FField.DataType = ftMemo) else
        FData := FFieldData;
      if VarIsArray(FData) then
        Result := VarArrayLock(FData) else
        Result := TVarData(FData).VOleStr;
    end;
  end;
end;

function TWideStringADOBlobStream.Write(const Buffer; Count: Longint): Longint;
begin
  Result := inherited Write(Buffer, Count);
  FModified := True;
end;

procedure TWideStringADOBlobStream.Truncate;
begin
  Clear;
  FModified := True;
end;

{ TWBADODataSet }

function TWBADODataSet.CreateBlobStream(Field: TField; Mode: TBlobStreamMode): TStream;
begin
  if Field is TWideADOMemoField then
    Result := TWideStringADOBlobStream.Create(Field as TBlobField, Mode)
  else Result := inherited CreateBlobStream(Field, Mode);
end;

type PFieldClass = ^TFieldClass;
initialization
//  RegisterFields([TWideADOMemoField]);
  PFieldClass(@DefaultFieldClasses[ftMemo])^ := TWideADOMemoField;
end.
